{$A+,B-,D+,E+,F-,G+,I+,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+}

(* CANVAS
	 ------
(C) 1998,99 by Arsne von Wyss - arsene@vonwyss.ch

Rev: 20000213

This source code is freeware for non-commercial use and provided AS-IS.
You are not allowed to use or spread incomplete (inclusive this header)
copies of this code. Use at your own risk. If you do not agree with these
statements, delete this source code immediately. Please contact the
author via email for questions or bug reports.

Source code in development and subject to change frequently. No support from
the author yet.*)

{The Canvas unit provides the graphics drawing facilities and the interface
for the graphics output drivers. Usually you'll not directy access a ~TCanvas~
but one of its descendants.}

unit Canvas;

(* Unit interface is HelpGen enabled *)

interface

uses
	Objects,Surfaces,Events,CanvasFX;

const
	{Colors are defined as 32-Bit integer value: $00RRGGBB

	You can use the constants to access the values easily, but using direct
	color codes will not affect compatibility in any way.}
	clBlack       =$000000;
	clBlue        =$000077;
	clGreen       =$007700;
	clCyan        =$007777;
	clRed         =$770000;
	clMagenta     =$770077;
	clBrown       =$777700;
	clLightGray   =$AAAAAA;
	clDarkGray    =$555555;
	clLightBlue   =$0000FF;
	clLightGreen  =$00FF00;
	clLightCyan   =$00FFFF;
	clLightRed    =$FF0000;
	clLightMagenta=$FF00FF;
	clYellow      =$FFFF00;
	clWhite       =$FFFFFF;
	clNone        =-1;
	{The line styles only apply to Line, Box and Circle procedures.}
	lsFull=[0..7];
	lsDot=[0,2,4,6];
	lsNegDot=lsFull-lsDot;
	lsDash=[0,1,2,3];
	lsNegDash=lsFull-lsDash;

type
	{For upwards compatibility only.}
	Cardinal=Word;
	{Text alignment, relative to the cursor. Also works for multi-line text.
	It is used in the ~Alignment~ property of the ~TCanvas~ object.}
	TAlignment=(taLeft,taCentered,taRight);
	{Line style, see ~lsXXXX~ constants.}
	TLineStyle=set of 0..7;
	{Color, see ~clXXXX~ constants}
	TColor=Longint;
	PFont=^TFont;
	PBitmap=^TBitmap;
	PCanvas=^TCanvas;
	{The canvas object is the graphics interface between all components of the GUI
	and the driver. A driver needs to inherit some of the functionaliy in order to
	make the TCanvas work. Note that there can be canvas for other purposes than
	screen output, like ~TSurfaceCanvas~.}
	TCanvas=object(TSurface)
	private
		UpdateCnt: Integer;
		LineStylePos: Integer;
	public
		{Read the color as a $RRGGBB-~TColor~-Value. This color value is not
		driver-dependent. Look at ~SetColor~ and the ~clXXXX~-values for more info.}
		GetColor: TColor;
		Alpha: Byte;
		{Read-Only. Offset relaive to top-level canvas; this is usually the screen.}
		Offset,
		{Position of the text cursor. You may modify it directly or using ~GotoXY~.}
		Cursor: TPoint;
		{The text "file" may be used instead of ~WriteString~ in relation with the
		systems Write and WriteLn functions, allowing to use the full flexibility
		of these functions.}
		Text: Text;
		{Font of the current canvas. You can change the font, however make sure it
		always points to a valid font!}
		Font: PFont;
		{Text alignment used for text output via both the ~text~ file variable and
		the ~WriteString~ function; it also respects multi-line text data. However,
		taCentered will only center horizontally.}
		Alignment: TAlignment;
		{LineStyle is only used by ~Line~, ~PolyLine~, ~Box~ and ~Circle~.}
		LineStyle: TLineStyle;
		{Canvas constructor. It should be called when creating a new canvas, however
		in most of the cases you'll not need to call a constructor yourself.}
		constructor Init;
		procedure DirectPutPixel(X,Y: Integer); virtual;
		procedure DirectHLine(X1,Y,X2: Integer); virtual; (* abstract *)
		procedure DirectVLine(X,Y1,Y2: Integer); virtual;
		procedure DirectBar(X1,Y1,X2,Y2: Integer); virtual;
		procedure DirectBlit(X1,Y1,X2,Y2,X3,Y3: Integer); virtual;
		procedure DirectReadHLine(X1,Y,X2: Integer; var Data); virtual;
		procedure DirectWriteHLine(X1,Y,X2: Integer; const Data); virtual;
		procedure DirectWriteHLineKey(X1,Y,X2: Integer; const Data; Key: TColor); virtual;
		procedure DirectPutPixelAlpha(X,Y: Integer); virtual;
		procedure DirectHLineAlpha(X1,Y,X2: Integer); virtual;
		procedure DirectVLineAlpha(X,Y1,Y2: Integer); virtual;
		procedure DirectBarAlpha(X1,Y1,X2,Y2: Integer); virtual;
		procedure DirectWriteHLineAlpha(X1,Y,X2: Integer; const Data); virtual;
{		procedure DirectWriteHLineKeyAlpha(X1,Y,X2: Integer; const Data; Key: TColor); virtual;}
		function DirectCalcHLine(X1,X2: Integer): Integer; virtual;
		function DirectGetPixel(X,Y: Integer): TColor; virtual; (* abstract *)

		{Draws an empty ellipse on the screen, using (X,Y) as centre and (R1,R2) for
		the axis radius. Rot allows you to rotate the ellipse.}
		procedure Ellipse(X,Y,R1,R2: Integer; Rot: Double);
		{Draws a filled ellipse on the screen, using (X,Y) as centre and (R1,R2) for
		the axis radius. Rot allows you to rotate the ellipse.}
		procedure FillEllipse(X,Y,R1,R2: Integer; Rot: Double);
		{Draws an bezier line on the screen, using (X1,Y1) and (X2,Y2) as fixed points,
		while (XA1,YA1) and (XA2,YA2) are the anchors.}
		procedure BezierLine(X1,Y1,XA1,YA1,XA2,YA2,X2,Y2: Integer);
		{Draws a series of bezier lines. The data format is as follows, in Integers:

		 (X1,Y1),(XA1,YA1), (XA2,YA2),(X2,Y2),(XB2,YB2), (XA3,YA3),(X3,Y3)
		 ?1: start point and one anchor
		 ?2: middle point and two anchors
		 ?3: end point and one anchor

		You can use as many middle points as you want (only limited by memory).}
		procedure PolyBezierLine(const PolyB; CountB: Integer);
		{Draws a filles shape limited of bezier lines. The data format is as
		follows, in Integers:

		 (X1,Y1),(XA1,YA1), (XA2,YA2),(X2,Y2),(XB2,YB2), (XA3,YA3),(X3,Y3)
		 ?1: start point and one anchor
		 ?2: middle point and two anchors
		 ?3: end point and one anchor

		You can use as many middle points as you want (limited by memory and bezier
		complexity). If the shape is not closed, a straight line will be used to
		close it.}
		procedure PolyBezierFill(const PolyB; CountB: Integer);
		{Puts a pixel on the canvas surface.}
		procedure PutPixel(X,Y: Integer);
		{Draws a horizontal line on the canvas surface from the position (X1,Y) to
		(X2,Y) with respect to the clipping area. You may set the drawing color
		first using ~SetColor~. Note that ~LineStyle~ is ignored in HLine.}
		procedure HLine(X1,Y,X2: Integer);
		{Draws a vertical line on the canvas surface from the position (X,Y1) to
		(X,Y2) with respect to the clipping area. You may set the drawing color
		first using ~SetColor~. Note that ~LineStyle~ is ignored in VLine.}
		procedure VLine(X,Y1,Y2: Integer);
		{Draws a line on the canvas surface from the position (X1,Y1) to
		(X2,Y2) with respect to the clipping area. You may set the drawing color
		using ~SetColor~ or a line style using ~LineStyle~.

		It is suggested to reset ~LineStyle~ back to a full line (see ~lsXXXX~
		constants).}
		procedure Line(X1,Y1,X2,Y2: Integer); virtual;
		{Draws a filled box on the canvas surface from the position (X1,Y1) to
		(X2,Y2) with respect to the clipping area. You may set the drawing color
		using ~SetColor~ first. The fill pattern cannot be altered (yet).}
		procedure Bar(X1,Y1,X2,Y2: Integer);
		{Draws a box on the canvas surface from the position (X1,Y1) to
		(X2,Y2) with respect to the clipping area. You may set the drawing color
		using ~SetColor~ or a line style using ~LineStyle~.

		It is suggested to reset ~LineStyle~ back to a full line (see ~lsXXXX~
		constants).}
		procedure Box(X1,Y1,X2,Y2: Integer); virtual;
		{Draws a circle on the canvas surface with the centre at (X,Y) and the
		radius R with respect to the clipping area. You may set the drawing color
		using ~SetColor~ or a line style using ~LineStyle~.

		It is suggested to reset ~LineStyle~ back to a full line (see ~lsXXXX~
		constants).}
		procedure Circle(X,Y,R: Integer); virtual;
		{Draws a filled circle on the canvas surface with the centre at (X,Y) and the
		radius R with respect to the clipping area. You may set the drawing color
		using ~SetColor~. The fill pattern pattern cannot be altered (yet).}
		procedure Disc(X,Y,R: Integer); virtual;
		{Low-level block move. Since the driver does not check any counds or clipping
		limits, you should not use Blit unless you know exactly what you are doing.
		THere will be a Copy()-function for clipped blits available in one of the
		next releases.}
		procedure	Blit(X1,Y1,X2,Y2,X3,Y3: Integer);
		{Draws a closed line polygon given by the TPoint-array and the point Count
		with respect to the clipping area. You may set the drawing color using
		~SetColor~ or a line style using ~LineStyle~.

		It is suggested to reset ~LineStyle~ back to a full line (see ~lsXXXX~
		constants).}
		procedure PolyLine(const Data; Count: Integer); virtual;
		{Draws a filled polygon given by the TPoint-array and the point Count
		with respect to the clipping area. You may set the drawing color using
		~SetColor~. The fill pattern pattern cannot be altered (yet).}
		procedure PolyFill(const Data; Count: Integer); virtual;
		{Tranfers the image block to the given stream. Note that you should use the
		stream types provided by the driver unit unless you want to store or load
		data from/to a specific stream.}
		procedure GetImage(Data: PBitmap; X1,Y1,X2,Y2: Integer); virtual; {abstract}
		{Draw bitmap.}
		procedure PutImage(Data: PBitmap; X,Y: Integer); virtual;
		{Draw shaded bitmap.}
		procedure PutImageShaded(Data: PBitmap; X,Y: Integer); virtual;
		{Draw tiled bitmap.}
		procedure PutImageShadedTiled(Data: PBitmap; X,Y: Integer; XTile,YTile: Word);
		{Draw tiled bitmap.}
		procedure PutImageTiled(Data: PBitmap; X,Y: Integer; XTile,YTile: Word);
		{Draws a 3D-looking box. If the border width is negative the left-top and
		right-bottom colors will be swapped. The canvas color will be set to
		CMiddle when the method terminates. Note that line styles are ignored.}
		procedure Box3D(X1,Y1,X2,Y2: Integer; CLeftTop,CMiddle,CRightBottom: TColor; AWidth: Integer);
		{Draws a filled 3D-looking box. If the border width is negative the left-top
		and right-bottom colors will be swapped. The canvas color will be set to
		CMiddle when the method terminates.}
		procedure Bar3D(X1,Y1,X2,Y2: Integer; CLeftTop,CMiddle,CRightBottom: TColor; AWidth: Integer);
		{Draws a 3D-looking circle witht he center (X,Y) and the radius R. A negative
		R will swap the left-top and right-bottom colors. Note that line styles are
		ignored.}
		procedure Circle3D(X,Y,R: Integer; CLeftTop,CRightBottom: TColor);
		{Clears the canvas with the current color. This does the same as drawing a
		~Bar~ over the whole canvas surface.}
		procedure Clear; virtual;
		{Clears the canvas with the specified color. This does the same as ~Clear~,
		but with the possibility to choose a different color. The color setting of
		the canvas will be reset to the value prior to to ClearColor.}
		procedure ClearColor(AColor: TColor);
		{Clears with bitmap or color}
		procedure ClearSmart(AColor: TColor; ABitmap: PBitmap; X,Y: Integer);
		{Locate text ~cursor~ at (X,Y).}
		procedure GotoXY(X,Y: Integer);
		{Sets the current color of the canvas. Note that the color format is alwayws
		$RRGGBB as described in the ~clXXXX~-constants.}
		procedure SetAlpha(NewAlpha: Byte);
		procedure SetColor(NewColor: TColor); virtual;
		{Notify the canvas object of upcoming drawing operations. BeginUpdate can
		enhance the canvas drawing performance because it will take care of nested
		calls and only ~Lock~ and ~Unlock~ the canvas once. Use ~EndUpdate~ when
		you're done with the drawing operations.}
		procedure BeginUpdate;
		{Notify the canvas of the finished update. See ~BeginUpdate~ for details.}
		procedure EndUpdate;
		{Lock the canvas. Will be called on the first ~BeginUpdate~ of the canvas,
		allowing the canvas descendant to take the actions necessary to allow drawing
		on the canvas. Lock should never be called directly but only be overridden
		by TCanvas-descendants.}
		procedure Lock; virtual;
		{Unlock the canvas. Will be called on the last ~EndUpdate~ of the canvas,
		allowing the canvas descendant to restore normal operation of the canvas
		and its parents. Unlock should never be called directly but only be overridden
		by TCanvas-descendants.}
		procedure UnLock; virtual;
(*		{Read the color as a $RRGGBB-~TColor~-Value. This color value is not
		driver-dependent. Look at ~SetColor~ and the ~clXXXX~-values for more info.}
		function GetColor: TColor; virtual; *)
		{Writes a string at the current ~cursor~ position using the current ~font~
		and text ~alignment~. The value returned will be the width in pixels of the
		written string.}
		function WriteString(const S: string): Integer;
		{Calculate the width of the string given using the current ~font~. Note that
		the most common use of this function, centering text, can be easily
		implemented by using the ~alignment~ property instead of computing the center
		by yourself.}
		function TextWidth(const S: string): Integer;
		{Calculate the height of the string. The height will be the height of the
		current ~font~ multiplied by the line changes in the string.}
		function TextHeight(const S: string): Integer;
		{Read pixel from the canvas surface. Note that this function may not be
		implemented by the driver, use with caution.}
		function GetPixel(X,Y: Integer): TColor; virtual; {abstract}
		{Tells wether one or more ~BeginUpdate~s have been called for this Canvas.}
		function Updating: Boolean;
	end;
	{The font anchestor object will never be directly used, but only through its
	descendants and the TCanvas unit. Don't care about the TFont unless you're
	implementing a driver.}
	TFont=object(TObject)
		{Returns the width of the character given.}
		function Width(C: Char): Integer; virtual;
		{Returns the height of the font.}
		function Height: Integer; virtual;
		{Writes the character to the given ~TCanvas~ and returns the width of the
		character written.}
		function Out(Canvas: PCanvas; C: Char): Integer; virtual;
	end;
	{Canvas to draw on surfaces. This one is handy if you want to modify a
	~TSurface~, for instance the shape of a ~TView~.

	As shown by the following code, it is not even necessary to have its
	own variable to use it:
	

		with PSurfaceCanvas(New(PSurfaceCanvas,Init(AnyView)))^ do begin
			SetColor(0);
			Bar(10,10,40,30); (* delete stuff *)
			SetColor(1);
			Bar(15,15,35,25); (* draw stuff *)
			Free;             (* release memory used by TSurfaceCanvas *)
		end;}
	PSurfaceCanvas=^TSurfaceCanvas;
	TSurfaceCanvas=object(TCanvas)
		Color: Boolean;
		{Drawing "color" on the surface. Since surfaces can be drawn or erased,
		this value is only a boolean.}
		Surface: PSurface;
		{Initialize the surface canvas. The parameter given is the surface you want
		to put the canvas on.}
		constructor Init(ASurface: PSurface);
		{Initialize the canvas to match the ~surface~. Note that you only need to
		call InitSurface when you changed ~surface~ or if the size of ~surface~
		changed.}
		procedure SetColor(AColor: TColor); virtual;
		{Sets the drawing color.}
		procedure InitSurface;
		{See ~TCanvas.PutPixel~ for details.}
		procedure DirectPutPixel(X,Y: Integer); virtual;
		{See ~TCanvas.HLine~ for details.}
		procedure DirectHLine(X1,Y,X2: Integer); virtual;
		{See ~TCanvas.VLine~ for details.}
		procedure DirectVLine(X,Y1,Y2: Integer); virtual;
	end;
	PTextBuf=^TTextBuf;
	{Text buffer used by ~TTextRec~}
	TTextBuf=array[0..127] of Char;
	{Only used internally for ~TCanvas.Text~ redirection.}
	TTextRec=record
		Handle: Word;
		Mode: Word;
		BufSize: Word;
		Private: Word;
		BufPos: Word;
		BufEnd: Word;
		BufPtr: PTextBuf;
		OpenFunc: Pointer;
		InOutFunc: Pointer;
		FlushFunc: Pointer;
		CloseFunc: Pointer;
		Canvas: PCanvas;
		UserData: array[5..16] of Byte;
		Name: array[0..78] of Char;
		Len: Byte;
		Buffer: TTextBuf;
	end;
	{Bitmap memory format. The possible balues are:

	bfNone:   The bitmap is not initialized
	bfMono:    1 Bit per pixel (8 Pixel per Bit)
	bf8:       8 Bit per pixel, RGB as     2:2:2(:L2)
	bf15:     15 Bit per pixel, RGB as     5:5:5
	bf16:     16 Bit per pixel, RGB as     5:6:5
	bf24:     24 Bit per pixel, RGB as     8:8:8
	bf32:     32 Bit per pixel, RGB as (8):8:8:8
	bfDriver: internal format of Driver, that is a "Device-Dependent Bitmap"}
	TBitmapFormat=Word;
	TBitmap=object(TReferenceObject)
		Width,Height,LineSize: Integer;
		Format: TBitmapFormat;
		InternalID: Longint;
		ColorKey: TColor;
		procedure Reset(AWidth,AHeight: Integer; AFormat: TBitmapFormat); virtual;
		procedure Convert(AFormat: TBitmapFormat); virtual;
		function EditLine(Y: Integer): Pointer; virtual;
		function ReadLine(Y: Integer): Pointer; virtual;
		procedure GetLine(Y: Integer; AData: Pointer); virtual;
		procedure SetLine(Y: Integer; AData: Pointer); virtual;
		procedure SaveToStream(AStream: PStream); virtual;
		procedure LoadFromStream(AStream: PStream); virtual;
		function Unique: Pointer; virtual;
		destructor Done; virtual;
	end;
	PPoint=^TPoint;
	PRect=^TRect;

const
	bfNone=0;
	bfRMask=$0F00;
	bfRShift=16;
	bfGMask=$00F0;
	bfGShift=8;
	bfBMask=$000F;
	bfBShift=0;
	bfDriver: TColor=$F000;

{Generate a random RGB color.}
function RandomColor: TColor;

{Can be used to quickly fill a TPoint structure:


	APoint=Point(20,30)^;}
function Point(AX,AY: Integer): PPoint;

{Can be used to quickly fill a TRect structure:


	ARect=Rect(40,30,100,110)^;}
function Rect(AX,AY,BX,BY: Integer): PRect;

implementation

uses
	DOS,GUI,GUIUtils,Bitmaps;

const
	Buffers=32;

type
	TFixed=record
		case Byte of
		0: (X: Longint);
		1: (Frac: Word; Int: Integer);
	end;
	PPointArray=^TPointArray;
	TPointArray=array[0..16381] of TPoint;
	TEdge=record
		X1,Y1,X2,Y2: Integer;
		CX,S: TFixed;
	end;
	PEdgeArray=^TEdgeArray;
	TEdgeArray=array[0..4094] of TEdge;
	PEdges=^TEdges;
	TEdges=object(TObject)
		Count,Alloc,Active,MinY,MaxY: Integer;
		Data,RealData: PEdgeArray;
		constructor Init(AAlloc: Integer);
		procedure Add(AX1,AY1,AX2,AY2: Integer);
		procedure Sort;
		procedure Activate(Y: Integer);
		procedure Remove(Nr: Integer);
		procedure Draw(Canvas: PCanvas);
		destructor Done; virtual;
	end;

var
	Points: record
		Act: Integer;
		Data: array[0..Buffers-1] of TPoint;
	end;
	Rects: record
		Act: Integer;
		Data: array[0..Buffers-1] of TRect;
	end;

function TextWrite(var Txt: TTextRec): Integer; far;
begin
	with Txt do begin
		Len:=BufPos;
		BufPos:=0;
		Canvas^.WriteString(PString(@Len)^);
	end;
	TextWrite:=0;
end;

function TextReturn(var Txt: TTextRec): Integer; far;
begin
	TextReturn:=0;
end;

function TextOpen(var Txt: TTextRec): Integer; far;
begin
	with Txt do
		if Mode<>fmOutput then
			TextOpen:=5 {File access denied}
		else begin
			InOutFunc:=@TextWrite;
			FlushFunc:=@TextWrite;
			CloseFunc:=@TextReturn;
			TextOpen:=0;
		end;
end;

constructor TCanvas.Init;
var
	T: TTextRec absolute Text;
begin
	inherited Init(0,0,False);
	with T do begin
		Mode:=fmClosed;
		BufSize:=SizeOf(Buffer);
		BufPtr:=@Buffer;
		OpenFunc:=@TextOpen;
		Name[0]:=#0;
		Canvas:=@self;
	end;
	Rewrite(Text);
	Font:=SystemFont;
	LineStyle:=lsFull;
end;

procedure TCanvas.DirectPutPixel(X,Y: Integer);
begin
	DirectHLine(X,Y,X);
end;

procedure TCanvas.DirectHLine(X1,Y,X2: Integer); (* abstract *)
begin
	Abstract;
end;

procedure TCanvas.DirectVLine(X,Y1,Y2: Integer);
var
	Y: Integer;
begin
	for Y:=Y1 to Y2 do
		DirectPutPixel(X,Y);
end;

procedure TCanvas.DirectBar(X1,Y1,X2,Y2: Integer);
var
	Y: Integer;
begin
	for Y:=Y1 to Y2 do
		DirectHLine(X1,Y,X2);
end;

procedure TCanvas.DirectBlit(X1,Y1,X2,Y2,X3,Y3: Integer); (* abstract *)
var
	Y,DiffX,AddY,DiffY,Size: Integer;
	Data: Pointer;
begin
	Size:=DirectCalcHLine(X1,X2);
	GetMem(Data,Size);
	DiffX:=X3-X1;
	DiffY:=Y3-Y1;
	AddY:=Sign(-DiffY) or 1;
	if AddY<0 then begin
		Y:=Y1;
		Y1:=Y2;
		Y2:=Y;
	end;
	Inc(Y2,AddY);
	Y:=Y1;
	repeat
		DirectReadHLine(X1,Y,X2,Data^);
		DirectWriteHLine(X3,Y+DiffY,X3-X1+X2,Data^);
		Inc(Y,AddY);
	until Y=Y2;
	FreeMem(Data,Size);
end;

function TCanvas.DirectGetPixel(X,Y: Integer): TColor;
var
	Src,Dst: Longint;
begin
	DirectReadHLine(X,Y,X,Src);
	BitmapProcessor^.ConvertFromScreen(Src,Dst,bf32,1);
	DirectGetPixel:=Dst;
end;

procedure TCanvas.Ellipse(X,Y,R1,R2: Integer; Rot: Double);
var
	R1X,R1Y,R2X,R2Y: Integer;
	Co,Si: Double;
	Bez: array[0..12] of TPoint;
begin
	Co:=Cos(Rot);
	Si:=Sin(Rot);
	R1X:=Round(Co*R1);
	R1Y:=Round(Si*R1);
	R2X:=-Round(Si*R2);
	R2Y:=Round(Co*R2);
	Bez[0].X:=X-R1X; {Start point}
	Bez[0].Y:=Y-R1Y;
	Bez[1].X:=Bez[0].X-(R2X div 2); {Anchor point}
	Bez[1].Y:=Bez[0].Y-(R2Y div 2);
	Bez[11].X:=Bez[0].X+(R2X div 2); {Anchor point}
	Bez[11].Y:=Bez[0].Y+(R2Y div 2);
	Bez[3].X:=X-R2X; {Start point}
	Bez[3].Y:=Y-R2Y;
	Bez[2].X:=Bez[3].X-(R1X div 2); {Anchor point}
	Bez[2].Y:=Bez[3].Y-(R1Y div 2);
	Bez[4].X:=Bez[3].X+(R1X div 2); {Anchor point}
	Bez[4].Y:=Bez[3].Y+(R1Y div 2);
	Bez[6].X:=X+R1X; {Start point}
	Bez[6].Y:=Y+R1Y;
	Bez[5].X:=Bez[6].X-(R2X div 2); {Anchor point}
	Bez[5].Y:=Bez[6].Y-(R2Y div 2);
	Bez[7].X:=Bez[6].X+(R2X div 2); {Anchor point}
	Bez[7].Y:=Bez[6].Y+(R2Y div 2);
	Bez[9].X:=X+R2X; {Start point}
	Bez[9].Y:=Y+R2Y;
	Bez[8].X:=Bez[9].X+(R1X div 2); {Anchor point}
	Bez[8].Y:=Bez[9].Y+(R1Y div 2);
	Bez[10].X:=Bez[9].X-(R1X div 2); {Anchor point}
	Bez[10].Y:=Bez[9].Y-(R1Y div 2);
	Bez[12]:=Bez[0];
	PolyBezierLine(Bez,5);
end;

procedure TCanvas.FillEllipse(X,Y,R1,R2: Integer; Rot: Double);
var
	R1X,R1Y,R2X,R2Y: Integer;
	Co,Si: Double;
	Bez: array[0..12] of TPoint;
begin
	Co:=Cos(Rot);
	Si:=Sin(Rot);
	R1X:=Round(Co*R1);
	R1Y:=Round(Si*R1);
	R2X:=-Round(Si*R2);
	R2Y:=Round(Co*R2);
	Bez[0].X:=X-R1X; {Start point}
	Bez[0].Y:=Y-R1Y;
	Bez[1].X:=Bez[0].X-(R2X div 2); {Anchor point}
	Bez[1].Y:=Bez[0].Y-(R2Y div 2);
	Bez[11].X:=Bez[0].X+(R2X div 2); {Anchor point}
	Bez[11].Y:=Bez[0].Y+(R2Y div 2);
	Bez[3].X:=X-R2X; {Start point}
	Bez[3].Y:=Y-R2Y;
	Bez[2].X:=Bez[3].X-(R1X div 2); {Anchor point}
	Bez[2].Y:=Bez[3].Y-(R1Y div 2);
	Bez[4].X:=Bez[3].X+(R1X div 2); {Anchor point}
	Bez[4].Y:=Bez[3].Y+(R1Y div 2);
	Bez[6].X:=X+R1X; {Start point}
	Bez[6].Y:=Y+R1Y;
	Bez[5].X:=Bez[6].X-(R2X div 2); {Anchor point}
	Bez[5].Y:=Bez[6].Y-(R2Y div 2);
	Bez[7].X:=Bez[6].X+(R2X div 2); {Anchor point}
	Bez[7].Y:=Bez[6].Y+(R2Y div 2);
	Bez[9].X:=X+R2X; {Start point}
	Bez[9].Y:=Y+R2Y;
	Bez[8].X:=Bez[9].X+(R1X div 2); {Anchor point}
	Bez[8].Y:=Bez[9].Y+(R1Y div 2);
	Bez[10].X:=Bez[9].X-(R1X div 2); {Anchor point}
	Bez[10].Y:=Bez[9].Y-(R1Y div 2);
	Bez[12]:=Bez[0];
	PolyBezierFill(Bez,5);
end;

procedure TCanvas.BezierLine(X1,Y1,XA1,YA1,XA2,YA2,X2,Y2: Integer);
var
	X,Y,OX,OY: Integer;
	AddMu,Mu,MuM1,MuM13,Mu3: Double;
begin
	AddMu:=20/Int(Sqrt(Sqr(Int(X1-XA1))+Sqr(Int(Y1-YA1)))+Sqrt(Sqr(Int(XA1-XA2))+Sqr(Int(YA1-YA2)))
			+Sqrt(Sqr(Int(XA2-X2))+Sqr(Int(YA2-Y2)))+60);
	Mu:=AddMu;
	OX:=X1;
	OY:=Y1;
	BeginUpdate;
	PutPixel(OX,OY);
	while Mu<1 do begin
		MuM1:=1-Mu;
		MuM13:=Sqr(MuM1)*MuM1;
		Mu3:=Sqr(Mu)*Mu;
		X:=Round(MuM13*X1+3*Mu*Sqr(MuM1)*XA1+3*Sqr(Mu)*MuM1*XA2+Mu3*X2);
		Y:=Round(MuM13*Y1+3*Mu*Sqr(MuM1)*YA1+3*Sqr(Mu)*MuM1*YA2+Mu3*Y2);
		Mu:=Mu+AddMu;
		case Abs(OX-X)+Abs(OY-Y) of
		0: Continue;
		1: PutPixel(X,Y);
		2: begin
				case Abs(OX-X) of
				0: PutPixel(X,Y+(OY-Y) div 2);
				2: PutPixel(X+(OX-X) div 2,Y);
				end;
				PutPixel(X,Y);
			end;
		else
			Line(OX,OY,X,Y);
		end;
		OX:=X;
		OY:=Y;
	end;
	if (OX<>X2) or (OY<>Y2) then
		Line(OX,OY,X2,Y2);
	EndUpdate;
end;

procedure TCanvas.PolyBezierLine(const PolyB; CountB: Integer);
var
	I: Integer;
	Points: TPointArray absolute PolyB;
begin
	if CountB=0 then
		Exit;
	if CountB=1 then begin
		PutPixel(Points[0].X,Points[0].Y);
		Exit;
	end;
	I:=0;
	CountB:=Pred(CountB)*3;
	BeginUpdate;
	while I<CountB do begin
		BezierLine(Points[I].X,Points[I].Y,Points[I+1].X,Points[I+1].Y,
							 Points[I+2].X,Points[I+2].Y,Points[I+3].X,Points[I+3].Y);
		Inc(I,3);
	end;
	EndUpdate;
end;

procedure TCanvas.PolyBezierFill(const PolyB; CountB: Integer);
const
	MaxComplexity=64;
var
	Points: TPointArray absolute PolyB;
	Edges: TEdges;
	OX,OY,I,Y: Integer;
	procedure CalcBezier(X1,Y1,XA1,YA1,XA2,YA2,X2,Y2: Integer);
	var
		X,Y: Integer;
		AddMu,Mu,MuM1,MuM13,Mu3: Double;
	begin
		AddMu:=20/Int(Sqrt(Sqr(Int(X1-XA1))+Sqr(Int(Y1-YA1)))+Sqrt(Sqr(Int(XA1-XA2))+Sqr(Int(YA1-YA2)))
				+Sqrt(Sqr(Int(XA2-X2))+Sqr(Int(YA2-Y2)))+40);
		if AddMu<=(1/MaxComplexity) then
			AddMu:=(1/(MaxComplexity-1));
		Mu:=0;
		while Mu<1 do begin
			MuM1:=1-Mu;
			MuM13:=Sqr(MuM1)*MuM1;
			Mu3:=Sqr(Mu)*Mu;
			X:=Round(MuM13*X1+3*Mu*Sqr(MuM1)*XA1+3*Sqr(Mu)*MuM1*XA2+Mu3*X2);
			Y:=Round(MuM13*Y1+3*Mu*Sqr(MuM1)*YA1+3*Sqr(Mu)*MuM1*YA2+Mu3*Y2);
			Mu:=Mu+AddMu;
			Edges.Add(OX,OY,X,Y);
			OX:=X;
			OY:=Y;
		end;
	end;
begin
	if CountB<2 then begin
		PolyLine(PolyB,CountB);
		Exit;
	end;
	Edges.Init(CountB*MaxComplexity+1);
	OX:=Points[0].X;
	OY:=Points[0].Y;
	I:=0;
	while I<Pred(CountB)*3 do begin
		CalcBezier(Points[I].X,Points[I].Y,Points[I+1].X,Points[I+1].Y,Points[I+2].X,Points[I+2].Y,Points[I+3].X,Points[I+3].Y);
		Inc(I,3);
	end;
	Edges.Add(OX,OY,Points[0].X,Points[0].Y);
	Edges.Draw(@self);
	Edges.Done;
end;

procedure TCanvas.PutPixel(X,Y: Integer);
begin
	if PointIncluded(X,Y) then begin
		BeginUpdate;
		if Alpha=0 then
			DirectPutPixel(X+Offset.X,Y+Offset.Y)
		else
			DirectPutPixelAlpha(X+Offset.X,Y+Offset.Y);
		EndUpdate;
	end;
end;

procedure TCanvas.HLine(X1,Y,X2: Integer);
var
	AHLine: THLine;
begin
	AHLine.A.X:=X1;
	AHLine.A.Y:=Y;
	AHLine.B.X:=X2;
	SurfaceResetHLine(AHLine);
	if SurfaceGetHLine(AHLine,X1,X2) then begin
		BeginUpdate;
		if Alpha=0 then
			repeat
				DirectHLine(X1+Offset.X,Y+Offset.Y,X2+Offset.X);
			until not SurfaceGetHLine(AHLine,X1,X2)
		else
			repeat
				DirectHLineAlpha(X1+Offset.X,Y+Offset.Y,X2+Offset.X);
			until not SurfaceGetHLine(AHLine,X1,X2);
		EndUpdate;
	end;
end;

procedure TCanvas.VLine(X,Y1,Y2: Integer);
var
	Y: Integer absolute Y1;
	LastY: Integer;
	Drawn: Boolean;
begin
	if Y1>Y2 then begin
		Y:=Y1;
		Y1:=Y2;
		Y2:=Y;
	end;
	LastY:=Y;
	Drawn:=False;
	while Y<=Y2 do begin
		if not PointIncluded(X,Y) then begin
			if LastY<Y then begin
				if not Drawn then begin
					Drawn:=True;
					BeginUpdate;
				end;
				if Alpha=0 then
					DirectVLine(X+Offset.X,LastY+Offset.Y,Y+Offset.Y-1)
				else
					DirectVLineAlpha(X+Offset.X,LastY+Offset.Y,Y+Offset.Y-1);
			end;
			LastY:=Y+1;
		end;
		Inc(Y);
	end;
	if LastY<Y then begin
		if not Drawn then begin
			Drawn:=True;
			BeginUpdate;
		end;
		if Alpha=0 then
			DirectVLine(X+Offset.X,LastY+Offset.Y,Y+Offset.Y-1)
		else
			DirectVLineAlpha(X+Offset.X,LastY+Offset.Y,Y+Offset.Y-1);
	end;
	if Drawn then
		EndUpdate;
end;

procedure TCanvas.Line(X1,Y1,X2,Y2: Integer);
var
	X,Y,LS: Integer;
	CX,AX: TFixed;
	Style: TLineStyle;
begin
	BeginUpdate;
	if LineStyle=lsFull then
		if Abs(X1-X2)>Abs(Y1-Y2) then begin
			if X1>X2 then begin
				X:=X1;
				Y:=Y1;
				X1:=X2;
				Y1:=Y2;
				X2:=X;
				Y2:=Y;
			end;
			AX.Frac:=0;
			AX.Int:=Y2-Y1;
			if X1<>X2 then
				AX.X:=AX.X div Succ(X2-X1);
			CX.Frac:=$7FFF;
			CX.Int:=Y1;
			for X:=X1 to X2 do begin
				Y:=CX.Int;
				Inc(CX.X,AX.X);
				VLine(X,Y,CX.Int-Sign(CX.Int-Y));
			end;
		end else begin
			if Y1>Y2 then begin
				X:=X1;
				Y:=Y1;
				X1:=X2;
				Y1:=Y2;
				X2:=X;
				Y2:=Y;
			end;
			AX.Frac:=0;
			AX.Int:=X2-X1;
			if Abs(X1-X2)=Abs(Y1-Y2) then
				AX.Int:=Sign(X2-X1)
			else if Y1<>Y2 then
				AX.X:=AX.X div Succ(Y2-Y1);
			CX.Frac:=$7FFF;
			CX.Int:=X1;
			for Y:=Y1 to Y2 do begin
				X:=CX.Int;
				Inc(CX.X,AX.X);
				HLine(X,Y,CX.Int-Sign(CX.Int-X));
			end;
		end
	else begin
		LS:=LineStylePos;
		Style:=LineStyle;
		if Abs(X1-X2)>Abs(Y1-Y2) then begin
			LineStylePos:=(LS+Abs(X2-X1)+1) and 7;
			if X1>X2 then begin
				X:=X1;
				Y:=Y1;
				X1:=X2;
				Y1:=Y2;
				X2:=X;
				Y2:=Y;
				LS:=(8-LineStylePos) and 7;
				Style:=lsFull-LineStyle;
			end;
			AX.Frac:=0;
			AX.Int:=Y2-Y1;
			if X1<>X2 then
				AX.X:=AX.X div Succ(X2-X1);
			CX.Frac:=$7FFF;
			CX.Int:=Y1;
			for X:=X1 to X2 do begin
				if LS in Style then
					PutPixel(X,CX.Int);
				Inc(CX.X,AX.X);
				LS:=Succ(LS) and 7;
			end;
		end else begin
			LineStylePos:=(LS+Abs(Y2-Y1)+1) and 7;
			if Y1>Y2 then begin
				X:=X1;
				Y:=Y1;
				X1:=X2;
				Y1:=Y2;
				X2:=X;
				Y2:=Y;
				LS:=(8-LineStylePos) and 7;
				Style:=lsFull-LineStyle;
			end;
			AX.Frac:=0;
			AX.Int:=X2-X1;
			if Abs(X1-X2)=Abs(Y1-Y2) then
				AX.Int:=Sign(X2-X1)
			else if Y1<>Y2 then
				AX.X:=AX.X div Succ(Y2-Y1);
			CX.Frac:=$7FFF;
			CX.Int:=X1;
			for Y:=Y1 to Y2 do begin
				if LS in Style then
					PutPixel(CX.Int,Y);
				Inc(CX.X,AX.X);
				LS:=Succ(LS) and 7;
			end;
		end;
	end;
	EndUpdate;
end;

procedure TCanvas.Bar(X1,Y1,X2,Y2: Integer);
var
	I: Integer;
	ToDraw: TSurface;
begin
	if X1>X2 then begin
		I:=X1;
		X1:=X2;
		X2:=I;
	end;
	if Y1>Y2 then begin
		I:=Y1;
		Y1:=Y2;
		Y2:=I;
	end;
	ToDraw.Init(Succ(X2-X1),Succ(Y2-Y1),False);
	ToDraw.SurfaceCombine(@self,-X1,-Y1,cmOr,False);
	if ToDraw.SurfaceCount>0 then begin
		BeginUpdate;
		if Alpha=0 then
			for I:=0 to ToDraw.SurfaceCount-1 do
				with ToDraw.SurfaceRect^[I] do
					DirectBar(X1+A.X+Offset.X,Y1+A.Y+Offset.Y,X1+B.X+Offset.X,Y1+B.Y+Offset.Y)
		else
			for I:=0 to ToDraw.SurfaceCount-1 do
				with ToDraw.SurfaceRect^[I] do
					DirectBarAlpha(X1+A.X+Offset.X,Y1+A.Y+Offset.Y,X1+B.X+Offset.X,Y1+B.Y+Offset.Y);
		EndUpdate;
	end;
	ToDraw.Done;
end;

procedure TCanvas.Box(X1,Y1,X2,Y2: Integer);
var
	I: Integer;
begin
	if Y1>Y2 then begin
		I:=Y1;
		Y1:=Y2;
		Y2:=I;
	end;
	BeginUpdate;
	if LineStyle=[0..7] then begin
		HLine(X1,Y1,X2);
		I:=Y2-Y1;
		if I>=1 then begin
			if I>=2 then begin
				VLine(X1,Y1+1,Y2-1);
				VLine(X2,Y1+1,Y2-1);
			end;
			HLine(X1,Y2,X2);
		end;
	end else begin
		if X1>X2 then begin
			I:=X1;
			X1:=X2;
			X2:=I;
		end;
		LineStylePos:=0;
		Line(X1,Y1,X2,Y1);
		I:=Y2-Y1;
		if I>=1 then begin
			if I>=2 then
				Line(X2,Y1+1,X2,Y2-1);
			Line(X2,Y2,X1,Y2);
			if I>=2 then
				Line(X1,Y2-1,X1,Y1+1);
		end;
	end;
	EndUpdate;
end;

procedure TCanvas.Circle(X,Y,R: Integer);
var
	D,CX,CY: Integer;
begin
	if R=0 then
		Exit;
	CY:=Abs(R);
	CX:=0;
	D:=CY-1;
	BeginUpdate;
	HLine(X+R,Y,X+R);
	HLine(X-R,Y,X-R);
	R:=0;
	{Add support for line styles!}
	repeat
		Dec(D,Succ(CX shl 1));
		Inc(CX);
		if D<0 then begin
			if R=0 then begin
				HLine(Succ(X-CX),Y+CY,Pred(X+CX));
				HLine(Succ(X-CX),Y-CY,Pred(X+CX));
			end else begin
				HLine(Succ(X-CX),Y+CY,X-R);
				HLine(Pred(X+CX),Y+CY,X+R);
				HLine(Succ(X-CX),Y-CY,X-R);
				HLine(Pred(X+CX),Y-CY,X+R);
			end;
			R:=CX;
			if CX=CY then
				Break;
			Dec(CY);
			Inc(D,CY shl 1);
		end;
		HLine(X+CY,Y+CX,X+CY);
		HLine(X+CY,Y-CX,X+CY);
		HLine(X-CY,Y+CX,X-CY);
		HLine(X-CY,Y-CX,X-CY);
	until CX>=CY;
	if (CX=CY) and (D=0) then begin {Correction for missing pixel error in rare cases}
		HLine(Succ(X-CX),Y+CY,X-R);
		HLine(Pred(X+CX),Y+CY,X+R);
		HLine(Succ(X-CX),Y-CY,X-R);
		HLine(Pred(X+CX),Y-CY,X+R);
	end;
	EndUpdate;
end;

procedure TCanvas.Disc(X,Y,R: Integer);
var
	D,CX,CY: Integer;
begin
	if R=0 then
		Exit;
	CY:=Abs(R);
	CX:=0;
	D:=CY-1;
	BeginUpdate;
	HLine(X-R,Y,X+R);
	repeat
		Dec(D,Succ(CX shl 1));
		if D<0 then begin
			HLine(X-CX,Y+CY,X+CX);
			HLine(X-CX,Y-CY,X+CX);
			if Succ(CX)=CY then
				Break;
			Dec(CY);
			Inc(D,CY shl 1);
		end;
		Inc(CX);
		HLine(X-CY,Y+CX,X+CY);
		HLine(X-CY,Y-CX,X+CY);
	until CX>=CY;
	EndUpdate;
end;

procedure TCanvas.DirectReadHLine(X1,Y,X2: Integer; var Data);
begin
	Abstract;
end;

procedure TCanvas.DirectWriteHLine(X1,Y,X2: Integer; const Data);
begin
	DirectWriteHLineKey(X1,Y,X2,Data,-1);
end;

procedure TCanvas.DirectWriteHLineKey(X1,Y,X2: Integer; const Data; Key: TColor);
begin
	Abstract;
end;

function TCanvas.DirectCalcHLine(X1,X2: Integer): Integer;
begin
	with TColorCode(Screen^.BitmapFormat) do
		DirectCalcHLine:=(Succ(X2-X1)*Bytes+3) and (not 3);
end;

procedure	TCanvas.Blit(X1,Y1,X2,Y2,X3,Y3: Integer);
begin
	BeginUpdate;
	{Some clipping should be added some day here}
	DirectBlit(X1+Offset.X,Y1+Offset.Y,X2+Offset.X,Y2+Offset.Y,X3+Offset.X,Y3+Offset.Y);
	EndUpdate;
end;

function TCanvas.Updating: Boolean;
begin
	Updating:=UpdateCnt>0;
end;

procedure TCanvas.PolyLine(const Data; Count: Integer);
var
	Points: TPointArray absolute Data;
begin
	if Count<=0 then
		Exit;
	case Count of
	1: with Points[0] do
			PutPixel(X,Y);
	2: Line(Points[0].X,Points[0].Y,Points[1].X,Points[1].Y);
	else
		Dec(Count);
		BeginUpdate;
		Line(Points[0].X,Points[0].Y,Points[Count].X,Points[Count].Y);
		while Count>0 do begin
			Line(Points[Count-1].X,Points[Count-1].Y,Points[Count].X,Points[Count].Y);
			Dec(Count);
		end;
		EndUpdate;
	end;
end;

procedure TCanvas.PolyFill(const Data; Count: Integer);
var
	Points: TPointArray absolute Data;
	Edges: TEdges;
	X1,X2,I,Y,HPos: Integer;
	AHLine: THLine;
begin
	if Count<3 then begin
		PolyLine(Data,Count);
		Exit;
	end;
	Edges.Init(Count);
	Dec(Count);
	for I:=0 to Count-1 do
		Edges.Add(Points[I].X,Points[I].Y,Points[I+1].X,Points[I+1].Y);
	Edges.Add(Points[Count].X,Points[Count].Y,Points[0].X,Points[0].Y);
	Edges.Draw(@self);
	Edges.Done;
end;

procedure TCanvas.GetImage(Data: PBitmap; X1,Y1,X2,Y2: Integer);
type
	PImgArray=^TImgArray;
	TImgArray=array[0..1599] of Word;
var
	I,X,Y: Integer;
	H: THLine;
	L: PImgArray;
begin
	Y:=Minimum(Y1,Y2);
	X:=Minimum(X1,X2);
	Data^.Reset(Succ(Abs(X2-X1)),Succ(Abs(Y2-Y1)),Screen^.BitmapFormat);
	if not RectOversect(X,Y,Pred(X+Data^.Width),Pred(Y+Data^.Height)) then
		Exit;
	BeginUpdate;
	for I:=0 to Data^.Height-1 do begin
		H.A.X:=X;
		H.A.Y:=Y+I;
		H.B.X:=X+Data^.Width-1;
		SurfaceResetHLine(H);
		L:=Data^.ReadLine(I);
		while SurfaceGetHLine(H,X1,X2) do
			DirectReadHLine(X1+Offset.X,Y+Offset.Y,X2+Offset.X,L^[X1-X]);
	end;
	EndUpdate;
end;

procedure TCanvas.PutImageShaded(Data: PBitmap; X,Y: Integer);
type
	PImgArray=^TImgArray;
	TImgArray=array[0..32768] of Byte;
var
	I,J,X1,X2: Integer;
	H: THLine;
	L: PImgArray;
	Temp: Pointer;
	TempSize: Word;
begin
	if Alpha=0 then
		PutImage(Data,X,Y);
	if Data^.Format<>Screen^.BitmapFormat then
		RunError;
	if not RectOversect(X,Y,Pred(X+Data^.Width),Pred(Y+Data^.Height)) then
		Exit;
	BeginUpdate;
	H.A.Y:=Y;
	TempSize:=DirectCalcHLine(X,X+Data^.Width-1);
	GetMem(Temp,TempSize);
	Inc(Y,Offset.Y);
	J:=TColorCode(Data^.Format).Bytes;
	for I:=0 to Data^.Height-1 do begin
		H.A.X:=X;
		H.B.X:=X+Data^.Width-1;
		SurfaceResetHLine(H);
		L:=Data^.ReadLine(I);
		while SurfaceGetHLine(H,X1,X2) do begin
			Move(L^[(X1-X)*J],Temp^,Succ(X2-X1)*J);
			BitmapProcessor^.AlphaFill(Temp^,Data^.Format,Succ(X2-X1),GetColor);
			DirectWriteHLine(X1+Offset.X,Y,X2+Offset.X,Temp^);
		end;
		Inc(H.A.Y);
		Inc(Y);
	end;
	FreeMem(Temp,TempSize);
	EndUpdate;
end;

procedure TCanvas.PutImageShadedTiled(Data: PBitmap; X,Y: Integer; XTile,YTile: Word);
var
	XC,YC: Integer;
begin
	if X<=-Data^.Width then begin
		XC:=(-X) div Data^.Width;
		Inc(X,XC*Data^.Width);
	end;
	if Y<=-Data^.Height then begin
		YC:=(-Y) div Data^.Height;
		Inc(Y,YC*Data^.Height);
	end;
	XTile:=Minimum((Width-X-1) div Data^.Width+1,XTile and $7FFF);
	YTile:=Minimum((Height-Y-1) div Data^.Height+1,YTile and $7FFF);
	if (XTile<=0) or (YTile<=0) then
		Exit;
	for YC:=0 to Pred(YTile) do
		for XC:=0 to Pred(XTile) do
			PutImageShaded(Data,X+Data^.Width*XC,Y+Data^.Height*YC);
end;

procedure TCanvas.PutImageTiled(Data: PBitmap; X,Y: Integer; XTile,YTile: Word);
var
	XC,YC: Integer;
begin
	if X<=-Data^.Width then begin
		XC:=(-X) div Data^.Width;
		Inc(X,XC*Data^.Width);
	end;
	if Y<=-Data^.Height then begin
		YC:=(-Y) div Data^.Height;
		Inc(Y,YC*Data^.Height);
	end;
	XTile:=Minimum((Width-X-1) div Data^.Width+1,XTile and $7FFF);
	YTile:=Minimum((Height-Y-1) div Data^.Height+1,YTile and $7FFF);
	if (XTile<=0) or (YTile<=0) then
		Exit;
	for YC:=0 to Pred(YTile) do
		for XC:=0 to Pred(XTile) do
			PutImage(Data,X+Data^.Width*XC,Y+Data^.Height*YC);
end;

procedure TCanvas.PutImage(Data: PBitmap; X,Y: Integer);
type
	PImgArray=^TImgArray;
	TImgArray=array[0..32768] of Byte;
var
	I,J,X1,X2: Integer;
	H: THLine;
	L: PImgArray;
	Key: TColor;
begin
	if Data^.Format<>Screen^.BitmapFormat then
		RunError;
	if not RectOversect(X,Y,Pred(X+Data^.Width),Pred(Y+Data^.Height)) then
		Exit;
	BeginUpdate;
	H.A.Y:=Y;
	Inc(Y,Offset.Y);
	J:=TColorCode(Data^.Format).Bytes;
	Key:=Data^.ColorKey;
	if Key=-1 then
		for I:=0 to Data^.Height-1 do begin
			H.A.X:=X;
			H.B.X:=X+Data^.Width-1;
			SurfaceResetHLine(H);
			L:=Data^.ReadLine(I);
			while SurfaceGetHLine(H,X1,X2) do
				DirectWriteHLine(X1+Offset.X,Y,X2+Offset.X,L^[(X1-X)*J]);
			Inc(H.A.Y);
			Inc(Y);
		end
	else
		for I:=0 to Data^.Height-1 do begin
			H.A.X:=X;
			H.B.X:=X+Data^.Width-1;
			SurfaceResetHLine(H);
			L:=Data^.ReadLine(I);
			while SurfaceGetHLine(H,X1,X2) do
				DirectWriteHLineKey(X1+Offset.X,Y,X2+Offset.X,L^[(X1-X)*J],Key);
			Inc(H.A.Y);
			Inc(Y);
		end;
	EndUpdate;
end;

procedure TCanvas.Box3D(X1,Y1,X2,Y2: Integer; CLeftTop,CMiddle,CRightBottom: TColor; AWidth: Integer);
var
	I: Integer;
begin
	if X1>X2 then begin
		I:=X1;
		X1:=X2;
		X2:=I;
	end;
	if Y1>Y2 then begin
		I:=Y1;
		Y1:=Y2;
		Y2:=I;
	end;
	if AWidth<0 then begin
		AWidth:=-AWidth;
		Swap(CLeftTop,CRightBottom,SizeOf(TColor));
	end;
	I:=Minimum(Succ(Abs(X2-X1)),Succ(Abs(Y2-Y1))) shr 1;
	if I<AWidth then
		AWidth:=I;
	BeginUpdate;
	if CLeftTop<>clNone then begin
		SetColor(CLeftTop);
		for I:=0 to AWidth-1 do begin
			HLine(X1+I,Y1+I,Pred(X2-I));
			VLine(X1+I,Succ(Y1+I),Pred(Y2-I));
		end;
	end;
	if CRightBottom<>clNone then begin
		SetColor(CRightBottom);
		for I:=0 to AWidth-1 do begin
			VLine(X2-I,Succ(Y1+I),Pred(Y2-I));
			HLine(Succ(X1+I),Y2-I,X2-I);
		end;
	end;
	if CMiddle<>clNone then begin
		SetColor(CMiddle);
		for I:=0 to AWidth-1 do begin
			PutPixel(X2-I,Y1+I);
			PutPixel(X1+I,Y2-I);
		end;
	end;
	EndUpdate;
end;

procedure TCanvas.Bar3D(X1,Y1,X2,Y2: Integer; CLeftTop,CMiddle,CRightBottom: TColor; AWidth: Integer);
var
	I: Integer;
begin
	if X1>X2 then begin
		I:=X1;
		X1:=X2;
		X2:=I;
	end;
	if Y1>Y2 then begin
		I:=Y1;
		Y1:=Y2;
		Y2:=I;
	end;
	if AWidth<0 then begin
		AWidth:=-AWidth;
		Swap(CLeftTop,CRightBottom,SizeOf(TColor));
	end;
	I:=Minimum(Succ(Abs(X2-X1)),Succ(Abs(Y2-Y1))) shr 1;
	if I<AWidth then
		AWidth:=I;
	BeginUpdate;
	Box3D(X1,Y1,X2,Y2,CLeftTop,CMiddle,CRightBottom,AWidth);
	if (CMiddle<>clNone) and (X2-X1>2*AWidth) and (Y2-Y1>2*AWidth) then
		Bar(X1+AWidth,Y1+AWidth,X2-AWidth,Y2-AWidth);
	EndUpdate;
end;

procedure TCanvas.Circle3D(X,Y,R: Integer; CLeftTop,CRightBottom: TColor);
var
	OldR,D,CX,CY: Integer;
begin
	if R=0 then
		Exit;
	if R<0 then begin
		R:=-R;
		Swap(CLeftTop,CRightBottom,4);
	end;
	SetColor(CLeftTop);
	OldR:=R;
	CY:=R;
	CX:=0;
	D:=CY-1;
	BeginUpdate;
	HLine(X-R,Y,X-R);
	R:=0;
	repeat
		Dec(D,Succ(CX shl 1));
		Inc(CX);
		if D<0 then begin
			if R=0 then begin
				HLine(Succ(X-CX),Y-CY,Pred(X+CX));
			end else begin
				HLine(Succ(X-CX),Y-CY,X-R);
				HLine(Pred(X+CX),Y-CY,X+R);
			end;
			R:=CX;
			if CX=CY then
				Break;
			Dec(CY);
			Inc(D,CY shl 1);
		end;
		HLine(X-CY,Y+CX,X-CY);
		HLine(X-CY,Y-CX,X-CY);
	until CX>=CY;
	if (CX=CY) and (D=0) then begin {Correction for missing pixel error in rare cases}
		HLine(Succ(X-CX),Y-CY,X-R);
		HLine(Pred(X+CX),Y-CY,X+R);
	end;
	SetColor(CRightBottom);
	CY:=OldR;
	CX:=0;
	D:=CY-1;
	HLine(X+OldR,Y,X+OldR);
	R:=0;
	repeat
		Dec(D,Succ(CX shl 1));
		Inc(CX);
		if D<0 then begin
			if R=0 then begin
				HLine(Succ(X-CX),Y+CY,Pred(X+CX));
			end else begin
				HLine(Succ(X-CX),Y+CY,X-R);
				HLine(Pred(X+CX),Y+CY,X+R);
			end;
			R:=CX;
			if CX=CY then
				Break;
			Dec(CY);
			Inc(D,CY shl 1);
		end;
		HLine(X+CY,Y+CX,X+CY);
		HLine(X+CY,Y-CX,X+CY);
	until CX>=CY;
	if (CX=CY) and (D=0) then begin {Correction for missing pixel error in rare cases}
		HLine(Succ(X-CX),Y+CY,X-R);
		HLine(Pred(X+CX),Y+CY,X+R);
	end;
	EndUpdate;
end;

procedure TCanvas.Clear;
begin
	Bar(0,0,Width-1,Height-1);
end;

procedure TCanvas.ClearColor(AColor: TColor);
var
	OldColor: TColor;
begin
	OldColor:=GetColor;
	SetColor(AColor);
	Clear;
	SetColor(OldColor);
end;

procedure TCanvas.ClearSmart(AColor: TColor; ABitmap: PBitmap; X,Y: Integer);
var
	OldColor: TColor;
begin
	if Assigned(ABitmap) then begin
		if AColor and $FF000000<>0 then begin
			OldColor:=GetColor;
			SetColor(AColor);
			PutImageShadedTiled(ABitmap,X,Y,$7FFF,$7FFF);
			SetColor(OldColor);
		end else
			PutImageTiled(ABitmap,X,Y,$7FFF,$7FFF);
	end else
		ClearColor(AColor and $FFFFFF);
end;

procedure TCanvas.DirectPutPixelAlpha(X,Y: Integer);
begin
	DirectHLineAlpha(X,Y,X);
end;

procedure TCanvas.DirectHLineAlpha(X1,Y,X2: Integer);
var
	Data: Pointer;
	Size: Word;
begin
	Size:=DirectCalcHLine(X1,X2);
	GetMem(Data,Size);
	DirectReadHLine(X1,Y,X2,Data^);
	BitmapProcessor^.AlphaFillScreen(Data^,Succ(X2-X1),GetColor);
	DirectWriteHLine(X1,Y,X2,Data^);
	FreeMem(Data,Size);
end;

procedure TCanvas.DirectVLineAlpha(X,Y1,Y2: Integer);
var
	Y: Integer;
begin
	for Y:=Y1 to Y2 do
		DirectPutPixelAlpha(X,Y);
end;

procedure TCanvas.DirectBarAlpha(X1,Y1,X2,Y2: Integer);
var
	Data: Pointer;
	Size: Word;
	Y: Integer;
begin
	Size:=DirectCalcHLine(X1,X2);
	GetMem(Data,Size);
	for Y:=Y1 to Y2 do begin
		DirectReadHLine(X1,Y,X2,Data^);
		BitmapProcessor^.AlphaFillScreen(Data^,Succ(X2-X1),GetColor);
		DirectWriteHLine(X1,Y,X2,Data^);
	end;
	FreeMem(Data,Size);
end;

procedure TCanvas.DirectWriteHLineAlpha(X1,Y,X2: Integer; const Data);
var
	Temp: Pointer;
	Size: Word;
begin
	Size:=DirectCalcHLine(X1,X2);
	GetMem(Temp,Size);
	DirectReadHLine(X1,Y,X2,Temp^);
	BitmapProcessor^.AlphaMixScreen(Data,Temp^,Succ(X2-X1),Alpha);
	DirectWriteHLine(X1,Y,X2,Temp^);
	FreeMem(Temp,Size);
end;

procedure TCanvas.BeginUpdate;
begin
	Inc(UpdateCnt);
	if UpdateCnt=1 then
		Lock;
end;

procedure TCanvas.EndUpdate;
begin
	Dec(UpdateCnt);
	if UpdateCnt=0 then
		UnLock;
end;

procedure TCanvas.Lock;
begin
	{nothing to do here}
end;

procedure TCanvas.UnLock;
begin
	{nothing to do here}
end;

procedure TCanvas.GotoXY(X,Y: Integer);
begin
	Cursor.X:=X;
	Cursor.Y:=Y;
end;

function TCanvas.WriteString(const S: string): Integer;
var
	I,J,X: Integer;
begin
	if not Assigned(Font) then begin
		WriteString:=0;
		Exit;
	end;
	BeginUpdate;
	X:=Cursor.X;
	case Alignment of
	taLeft: begin
			for I:=1 to Length(S) do
				case S[I] of
				#7: Write(Screen^.OriginalOutput,#7);
				#13: Cursor.X:=0;
				#10: Inc(Cursor.Y,Font^.Height);
				else
					Inc(Cursor.X,Font^.Out(@self,S[I]));
				end;
			WriteString:=Abs(Cursor.X-X);
		end;
	taCentered: begin
			J:=1;
			while (S[J]<>#13) and (J<=Length(S)) do
				Inc(J);
			Dec(Cursor.X,TextWidth(system.Copy(S,1,J-1)) shr 1);
			for I:=1 to Length(S) do
				case S[I] of
				#7: Write(Screen^.OriginalOutput,#7);
				#13: begin
						J:=I+1;
						while (S[J]<>#13) and (J<=Length(S)) do
							Inc(J);
						Cursor.X:=X-TextWidth(system.Copy(S,I+1,J-I)) shr 1;
					end;
				#10: Inc(Cursor.Y,Font^.Height);
				else
					Inc(Cursor.X,Font^.Out(@self,S[I]));
				end;
			Cursor.X:=X;
			WriteString:=Abs(Cursor.X-X);
		end;
	taRight: begin
			for I:=Length(S) downto 1 do
				case S[I] of
				#7: Write(Screen^.OriginalOutput,#7);
				#13: Cursor.X:=X;
				#10: Dec(Cursor.Y,Font^.Height);
				else
					Dec(Cursor.X,Font^.Width(S[I]));
					Font^.Out(@self,S[I]);
				end;
			WriteString:=Abs(Cursor.X-X);
		end;
	end;
	EndUpdate;
end;

function TCanvas.TextWidth(const S: string): Integer;
var
	I,X,MaxX: Integer;
begin
	if not Assigned(Font) then begin
		TextWidth:=0;
		Exit;
	end;
	X:=0;
	MaxX:=0;
	for I:=1 to Length(S) do
		case S[I] of
		#13: begin
				if X>MaxX then
					MaxX:=X;
				X:=0;
			end;
		else
			Inc(X,Font^.Width(S[I]));
		end;
	if MaxX>X then
		X:=MaxX;
	TextWidth:=X;
end;

function TCanvas.TextHeight(const S: string): Integer;
var
	I,H,Cnt: Integer;
begin
	if not Assigned(Font) then begin
		TextHeight:=0;
		Exit;
	end;
	H:=Font^.Height;
	Cnt:=H;
	for I:=1 to Length(S) do
		if S[I]=#10 then
			Inc(Cnt,H);
	TextHeight:=Cnt;
end;

function TCanvas.GetPixel(X,Y: Integer): TColor;
begin
	BeginUpdate;
	GetPixel:=DirectGetPixel(X+Offset.X,Y+Offset.Y);
	EndUpdate;
end;

procedure TCanvas.SetColor(NewColor: TColor);
var
	RGBA: packed record
		R,G,B,A: Byte;
	end absolute NewColor;
begin
	GetColor:=NewColor;
	Alpha:=RGBA.A;
end;

procedure TCanvas.SetAlpha(NewAlpha: Byte);
var
	RGBL: TColor;
	RGBA: packed record
		R,G,B,A: Byte;
	end absolute RGBL;
begin
	RGBL:=GetColor;
	RGBA.A:=NewAlpha;
	SetColor(RGBL);
end;

function TFont.Width(C: Char): Integer;
begin
	Abstract;
end;

function TFont.Height: Integer;
begin
	Abstract;
end;

function TFont.Out(Canvas: PCanvas; C: Char): Integer;
begin
	Abstract;
end;

constructor TSurfaceCanvas.Init(ASurface: PSurface);
begin
	inherited Init;
	Surface:=ASurface;
	InitSurface;
end;

procedure TSurfaceCanvas.InitSurface;
begin
	if Assigned(Surface) then begin
		with Surface^ do
			self.SurfaceResize(Width,Height,False);
		SurfaceClear(True);
	end;
end;

procedure TSurfaceCanvas.SetColor(AColor: TColor);
begin
	inherited SetColor(AColor);
	Color:=AColor<>0;
end;

procedure TSurfaceCanvas.DirectPutPixel(X,Y: Integer);
begin
	if Assigned(Surface) then
		Surface^.SurfaceBar(X,Y,X,Y,Color);
end;

procedure TSurfaceCanvas.DirectHLine(X1,Y,X2: Integer);
begin
	if Assigned(Surface) then
		Surface^.SurfaceBar(X1,Y,X2,Y,Color);
end;

procedure TSurfaceCanvas.DirectVLine(X,Y1,Y2: Integer);
begin
	if Assigned(Surface) then
		Surface^.SurfaceBar(X,Y1,X,Y2,Color);
end;

procedure TBitmap.Reset(AWidth,AHeight: Integer; AFormat: TBitmapFormat);
begin
	Abstract;
end;

procedure TBitmap.Convert(AFormat: TBitmapFormat);
begin
	Abstract;
end;

function TBitmap.EditLine(Y: Integer): Pointer;
begin
	Abstract;
end;

function TBitmap.ReadLine(Y: Integer): Pointer;
begin
	ReadLine:=EditLine(Y);
end;

procedure TBitmap.GetLine(Y: Integer; AData: Pointer);
begin
	Move(ReadLine(Y)^,AData^,LineSize);
end;

procedure TBitmap.SetLine(Y: Integer; AData: Pointer);
begin
	Move(AData^,EditLine(Y)^,LineSize);
end;

function TBitmap.Unique: Pointer;
begin
	Abstract;
end;

procedure TBitmap.SaveToStream(AStream: PStream);
var
	Y: Integer;
begin
	with AStream^ do begin
		Write(Width,SizeOf(Width));
		Write(Height,SizeOf(Height));
		Write(Format,SizeOf(Format));
		Write(InternalID,SizeOf(InternalID));
		for Y:=0 to Height-1 do
			Write(EditLine(Y)^,LineSize);
	end;
end;

procedure TBitmap.LoadFromStream(AStream: PStream);
var
	W,H,Y: Integer;
	F: TBitmapFormat;
begin
	with AStream^ do begin
		Read(W,SizeOf(Width));
		Read(H,SizeOf(Height));
		Read(F,SizeOf(Format));
		self.Reset(W,H,F);
		Read(InternalID,SizeOf(InternalID));
		for Y:=0 to Height-1 do
			Read(EditLine(Y)^,LineSize);
	end;
end;

destructor TBitmap.Done;
begin
	Reset(0,0,bfNone);
	inherited Done;
end;

constructor TEdges.Init(AAlloc: Integer);
begin
	inherited Init;
	Alloc:=AAlloc;
	if MaxAvail<Alloc*SizeOf(TEdge)-1024 then
		Alloc:=Maximum(4,(MaxAvail-1024) div SizeOf(TEdge));
	GetMem(Data,Alloc*SizeOf(TEdge));
	RealData:=Data;
	MinY:=$7FFF;
end;

procedure TEdges.Add(AX1,AY1,AX2,AY2: Integer);
var
	I: Integer;
begin
	if AY1=AY2 then
		Exit;
	if Count>=Alloc then
		{$IFDEF DEBUG}RunError;{$ELSE}Exit;{$ENDIF}
	with Data^[Count] do begin
		if AY1<=AY2 then begin
			X1:=AX1;
			Y1:=AY1;
			X2:=AX2;
			Y2:=AY2;
		end else begin
			X1:=AX2;
			Y1:=AY2;
			X2:=AX1;
			Y2:=AY1;
		end;
		if Y1<MinY then
			MinY:=Y1;
		if Y2>MaxY then
			MaxY:=Y2;
	end;
	Inc(Count);
end;

procedure TEdges.Sort; {Sort inactive edges, Quick Sort}
	procedure QSort(L,R: Integer);
	var
		I,J,Y: Integer;
	begin
		I:=L;
		J:=R;
		Y:=Data^[(L+R) shr 1].Y1;
		repeat
			while Data^[I].Y1<Y do
				Inc(I);
			while Y<Data^[J].Y1 do
				Dec(J);
			if I<=J then begin
				Swap(Data^[I],Data^[J],SizeOf(TEdge));
				Inc(I);
				Dec(J);
			end;
		until I>J;
		if L<J then
			QSort(L,J);
		if I<R then
			QSort(I,R);
	end;
begin
	if Count>1 then
		QSort(Active,Count-1);
end;

procedure TEdges.Activate(Y: Integer);
var
	I,J: Integer;
	Temp: TEdge;
begin
	while (Active<Count) and (Data^[Active].Y1<=Y) do begin
		with Data^[Active] do begin
			S.X:=((X2-X1)*65536) div (Y2-Y1);
			CX.Int:=X1;
			CX.Frac:=$7FFF;
			Dec(CX.X,Succ(Y1-Y)*S.X);
		end;
		Inc(Active);
	end;
	for I:=Pred(Active) downto 0 do
		with Data^[I] do
			if Y>=Data^[I].Y2 then
				Remove(I)
			else
				Inc(CX.X,S.X); {add slope}
	for I:=1 to Active-1 do begin {sort actives (insertion)}
		Temp:=Data^[I];
		J:=I;
		while Data^[J-1].CX.X>Temp.CX.X do begin
			Data^[J]:=Data^[J-1];
			Dec(J);
			if J=0 then
				Break;
		end;
		if J<>I then
			Data^[J]:=Temp;
	end;
end;

procedure TEdges.Remove(Nr: Integer);
begin
	Dec(Count);
	if Nr=0 then
		Inc(Longint(Data),SizeOf(TEdge))
	else if Nr<Count then
		Move(Data^[Nr+1],Data^[Nr],SizeOf(TEdge)*(Count-Nr));
	if Nr<Active then
		Dec(Active);
end;

procedure TEdges.Draw(Canvas: PCanvas); {Optimized filling of the line list}
var
	I,X1,X2,Y: Integer;
	AHLine: THLine;
begin
	with Canvas^ do begin
		Sort;
		if MinY<0 then
			MinY:=0;
		if MaxY>=Height then
			MaxY:=Height-1;
		BeginUpdate;
		for Y:=MinY to MaxY do begin
			Activate(Y);
			if Active>0 then begin
				AHLine.A.Y:=Y;
				for I:=0 to Active div 2-1 do begin
					AHLine.A.X:=Data^[I shl 1].CX.Int;
					AHLine.B.X:=Data^[I shl 1+1].CX.Int;
					SurfaceResetHLine(AHLine);
					if Alpha=0 then
						while SurfaceGetHLine(AHLine,X1,X2) do
							DirectHLine(X1+Offset.X,Y+Offset.Y,X2+Offset.X)
					else
						while SurfaceGetHLine(AHLine,X1,X2) do
							DirectHLineAlpha(X1+Offset.X,Y+Offset.Y,X2+Offset.X);
				end;
			end;
		end;
		EndUpdate;
	end;
end;

destructor TEdges.Done;
begin
	FreeMem(RealData,Alloc*SizeOf(TEdge));
	inherited Done;
end;

function Point(AX,AY: Integer): PPoint;
begin
	with Points do begin
		with Data[Act] do begin
			X:=AX;
			Y:=AY;
		end;
		Point:=@Data[Act];
		Inc(Act);
		if Act>=Buffers then
			Act:=0
	end;
end;

function Rect(AX,AY,BX,BY: Integer): PRect;
begin
	with Rects do begin
		with Data[Act] do begin
			A.X:=AX;
			A.Y:=AY;
			B.X:=BX;
			B.Y:=BY;
		end;
		Rect:=@Data[Act];
		Inc(Act);
		if Act>=Buffers then
			Act:=0
	end;
end;

function RandomColor: TColor;
begin
	RandomColor:=Random(256)+Random(256)*256+Random(256)*65536;
end;

end.